import json
from copy import deepcopy
from pathlib import Path
from typing import Any, Dict, TypedDict, Union

from typing_extensions import NotRequired

from ..config import ConfigError, load_class
from ..filestore.backends.local import LocalStorage
from .reader import OasisReader


class ResolvedReaderEngineConfig(TypedDict):
    path: str
    options: Dict[str, Any]


class ResolvedReaderConfig(TypedDict):
    filepath: str
    engine: ResolvedReaderEngineConfig


class InputReaderEngineConfig(TypedDict):
    path: NotRequired[str]
    options: NotRequired[Dict[str, Any]]


class InputReaderConfig(TypedDict):
    filepath: str
    engine: NotRequired[Union[str, InputReaderEngineConfig]]


def clean_config(config: Union[str, InputReaderConfig]) -> ResolvedReaderConfig:
    if isinstance(config, (str, Path)) or hasattr(config, "read"):
        _config: dict = {
            "filepath": config,
        }
    elif not isinstance(config, dict):
        raise ConfigError(f"df_reader config must be a string or dictionary: {config}")
    else:
        config: dict  # type: ignore
        _config = deepcopy(config)  # type: ignore

    if "filepath" not in _config:
        raise ConfigError(
            f"df_reader config must provide a 'filepath' property: {_config}"
        )

    if "engine" not in _config:
        _config["engine"] = {
            "path": "oasis_data_manager.df_reader.reader.OasisPandasReader",
            "options": {},
        }
    elif isinstance(_config.get("engine"), str):
        try:
            # try to decode the string a json object so it can be
            # serialized on the command line
            _config["engine"] = json.loads(_config.get("engine"))  # type: ignore
        except json.JSONDecodeError:
            _config["engine"] = {"path": _config.get("engine"), "options": {}}

    _config["engine"].setdefault("path", "oasis_data_manager.df_reader.reader.OasisPandasReader")
    _config["engine"].setdefault("options", {})

    return _config  # type: ignore


def get_df_reader(config, *args, **kwargs):
    config = clean_config(config)

    cls = load_class(config["engine"]["path"], OasisReader)

    storage = config["engine"]["options"].pop("storage", None) or LocalStorage("/")
    return cls(
        config["filepath"], storage, *args, **kwargs, **config["engine"]["options"]
    )



import importlib


class ConfigError(Exception):
    pass


def load_class(path, base=None):
    path_split = path.rsplit(".", 1)
    if len(path_split) != 2:
        raise ConfigError(f"'path' found in the df_reader config is not valid: {path}")

    module_path, cls_name = path_split
    module = importlib.import_module(module_path)
    cls = getattr(module, cls_name)

    if base and cls is not base and base not in cls.__bases__:
        raise ConfigError(f"'{cls.__name__}' does not extend '{base.__name__}'")

    return cls



import logging
import pathlib
from io import BytesIO
from typing import List, Type

import httpcore
import httpx
import pandas as pd

from oasis_data_manager.df_reader.config import InputReaderConfig, clean_config, get_df_reader
from oasis_data_manager.df_reader.reader import OasisReader
from oasis_data_manager.filestore.backends.local import LocalStorage

logger = logging.getLogger(__name__)


class Adjustment:
    """
    Adjustments are any Pandas adjustments made after the data is fetched and filtered by SQL if applicable.
    """

    @classmethod
    def apply(cls, df):
        return df


class ComplexData:
    adjustments: List[Type[Adjustment]] = []
    filename: str = ""
    url: str = ""
    fetch_required: bool = True
    sql: str = ""

    def __init__(self, storage=None):
        if not storage:
            storage = LocalStorage()
        self.storage = storage

    def fetch(self):
        raise NotImplementedError

    def get_sql(self):
        return self.sql

    def adjust(self, reader) -> OasisReader:
        """
        Hook to apply any adjustments.

        TODO adjustments are filters? Functions fun on the readers df i.e
        apply, should be change filter to apply then or is that confusing with pandas?
        """
        if self.adjustments:
            return reader.filter([a.apply for a in self.adjustments])
        return reader

    def to_dataframe(self, result) -> pd.DataFrame:
        """
        Hook to allow conversion of the fetch() to a dataframe, in the case of the
        RestComplexData class for example, the result is json, so can be fed
        directly to a dataframe and wrapped into our df_reader.
        """
        return pd.DataFrame(result)

    def get_df_reader(self, filepath, **kwargs) -> OasisReader:
        df_reader_config = clean_config(
            InputReaderConfig(
                filepath=filepath,
                engine="oasis_data_manager.df_reader.reader.OasisDaskReader",
            )
        )
        df_reader_config["engine"]["options"]["storage"] = self.storage

        return get_df_reader(df_reader_config, **kwargs)

    def to_reader(self, fetch_result) -> OasisReader:
        if fetch_result:
            # When the result has been fetched, apply to_dataframe and pass directly into the df_reader
            df = self.to_dataframe(fetch_result)
            return self.get_df_reader(
                None,  # TODO - None value instead of filename_or_buffer, improve this.
                dataframe=df,
                has_read=True,
            )
        else:
            # Not fetched, let df_reader fetch as normal.
            return self.get_df_reader(self.filename if self.filename else self.url)

    def run(self):
        if self.fetch_required and self.filename:
            filename_or_url = self.filename if self.filename else self.url
            extension = pathlib.Path(filename_or_url).suffix
            self.fetch_required = extension not in [".parquet", ".pq", ".csv"]

        fetch_result = None
        if self.fetch_required:
            fetch_result = self.fetch()

        reader = self.to_reader(fetch_result)

        sql = self.get_sql()
        if hasattr(reader, "sql") and sql:
            reader = reader.sql(sql)

        reader = self.adjust(reader)

        # TODO store file? return for now
        return reader


class FileStoreComplexData(ComplexData):
    def to_dataframe(self, result) -> pd.DataFrame:
        """
        As this is only called on filetypes not handled by the df_reader, this will always be custom.
        """
        raise NotImplementedError

    def fetch(self):
        with self.storage.open(self.filename, "rb") as f:
            result = BytesIO(f.read())
        return result


class RestComplexData(ComplexData):
    exceptions = (
        httpx.RequestError,
        httpx.TimeoutException,
        httpx.ReadTimeout,
        httpx.ConnectTimeout,
        httpx.ConnectError,
        httpcore.ReadTimeout,
        httpcore.ConnectTimeout,
        httpcore.ConnectError,
    )

    url: str
    timeout: int = 10

    def handle_error(self, exception, **kwargs):
        logger.warning(
            "Exception in complex data call",
            extra={"exception": exception, "uri": self.url},
        )
        return None

    def handle_response(self, response) -> dict:
        return response.json()

    def get_uri(self) -> str:
        return self.url

    def get_headers(self) -> dict:
        return {}

    def get_post_json(self) -> dict:
        return {}

    def fetch(self):
        uri = self.get_uri()
        headers = self.get_headers()
        post_json = self.get_post_json()
        timeout = self.timeout

        try:
            if post_json:
                response = httpx.post(
                    uri, json=post_json, headers=headers, timeout=timeout
                )
            else:
                response = httpx.get(uri, headers=headers, timeout=timeout)

            if response.status_code == 200:
                return self.handle_response(response)
            else:
                return self.handle_error(
                    exception="Unexpected status in complex data call",
                    post_json=post_json,
                )
        except self.exceptions as e:
            return self.handle_error(exception=str(e), post_json=post_json)



import logging


def set_aws_log_level(log_level):
    # Set log level for s3boto3
    try:
        LOG_LEVEL = getattr(logging, log_level.upper())
    except AttributeError:
        LOG_LEVEL = logging.WARNING

    logging.getLogger("boto3").setLevel(LOG_LEVEL)
    logging.getLogger("botocore").setLevel(LOG_LEVEL)
    logging.getLogger("nose").setLevel(LOG_LEVEL)
    logging.getLogger("s3transfer").setLevel(LOG_LEVEL)
    logging.getLogger("urllib3").setLevel(LOG_LEVEL)


def set_azure_log_level(log_level):
    try:
        LOG_LEVEL = getattr(logging, log_level.upper())
    except AttributeError:
        LOG_LEVEL = logging.WARNING
    logging.getLogger("azure").setLevel(LOG_LEVEL)



import contextlib
import os
from pathlib import Path
from typing import Optional
from urllib import parse
from urllib.parse import parse_qsl, urlsplit

import fsspec
from fsspec.asyn import sync

from ..log import set_aws_log_level
from .base import BaseStorage


class AwsS3Storage(BaseStorage):
    fsspec_filesystem_class = fsspec.get_filesystem_class("s3")

    def __init__(
        self,
        bucket_name: Optional[str] = None,
        access_key: Optional[str] = None,
        secret_key: Optional[str] = None,
        endpoint_url: Optional[str] = None,
        file_overwrite=True,
        object_parameters: Optional[dict] = None,
        auto_create_bucket=False,
        default_acl: Optional[str] = None,
        bucket_acl: Optional[str] = None,
        querystring_auth=False,
        querystring_expire=604800,
        reduced_redundancy=False,
        location="",
        encryption=False,
        security_token=None,
        secure_urls=True,
        file_name_charset="utf-8",
        gzip=False,
        preload_metadata=False,
        url_protocol="http:",
        region_name=None,
        use_ssl=True,
        verify=None,
        max_memory_size=0,
        shared_bucket=False,
        aws_log_level="",
        gzip_content_types=(
            "text/css",
            "text/javascript",
            "application/javascript",
            "application/x-javascript",
            "image/svg+xml",
        ),
        root_dir="",
        **kwargs,
    ):
        """Storage Connector for Amazon S3

        Store objects in a bucket common to a single worker pool. Returns a pre-signed URL
        as a response to the server which is downloaded and stored by Django-storage module

        Documentation
        -------------
        https://github.com/jschneier/django-storages/blob/master/storages/backends/s3boto3.py
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#id244

        TODO
        ----

        * Add optional local caching
        * option to set object expiry policy on bucket

            def _get_bucket_policy(self):
                pass
            def _set_lifecycle(self, ):
                pass
                https://stackoverflow.com/questions/14969273/s3-object-expiration-using-boto
                https://boto3.amazonaws.com/v1/documentation/api/latest/guide/s3-example-bucket-policies.html

        Parameters
        ----------
        :param settings_conf: Settings object for worker
        :type settings_conf: src.conf.iniconf.Settings
        """
        bucket_acl = bucket_acl or default_acl
        object_parameters = object_parameters or {}

        # Required
        self.storage_connector = "AWS-S3"
        self._bucket = None
        self._connection = None
        self.bucket_name = bucket_name

        # Optional
        self.access_key = access_key
        self.secret_key = secret_key
        self.endpoint_url = endpoint_url
        self.file_overwrite = file_overwrite
        self.object_parameters = object_parameters
        self.auto_create_bucket = auto_create_bucket
        self.default_acl = default_acl
        self.bucket_acl = bucket_acl
        self.querystring_auth = querystring_auth
        self.querystring_expire = querystring_expire
        self.reduced_redundancy = reduced_redundancy
        self.location = location
        self.encryption = encryption
        self.security_token = security_token
        self.secure_urls = secure_urls
        self.file_name_charset = file_name_charset
        self.gzip = gzip
        self.preload_metadata = preload_metadata
        self.url_protocol = url_protocol
        self.region_name = region_name
        self.use_ssl = use_ssl
        self.verify = verify
        self.max_memory_size = max_memory_size
        self.shared_bucket = shared_bucket
        self.aws_log_level = aws_log_level
        self.gzip_content_types = gzip_content_types
        set_aws_log_level(self.aws_log_level)

        root_dir = os.path.join(self.bucket_name or "", root_dir)
        if root_dir.startswith(os.path.sep):
            root_dir = root_dir[1:]
        if root_dir.endswith(os.path.sep):
            root_dir = root_dir[:-1]

        super(AwsS3Storage, self).__init__(root_dir=root_dir, **kwargs)

    @property
    def config_options(self):
        return {
            "bucket_name": self.bucket_name,
            "access_key": self.access_key,
            "secret_key": self.secret_key,
            "endpoint_url": self.endpoint_url,
            "file_overwrite": self.file_overwrite,
            "object_parameters": self.object_parameters,
            "auto_create_bucket": self.auto_create_bucket,
            "default_acl": self.default_acl,
            "bucket_acl": self.bucket_acl,
            "querystring_auth": self.querystring_auth,
            "querystring_expire": self.querystring_expire,
            "reduced_redundancy": self.reduced_redundancy,
            "location": self.location,
            "encryption": self.encryption,
            "security_token": self.security_token,
            "secure_urls": self.secure_urls,
            "file_name_charset": self.file_name_charset,
            "gzip": self.gzip,
            "preload_metadata": self.preload_metadata,
            "url_protocol": self.url_protocol,
            "region_name": self.region_name,
            "use_ssl": self.use_ssl,
            "verify": self.verify,
            "max_memory_size": self.max_memory_size,
            "shared_bucket": self.shared_bucket,
            "aws_log_level": self.aws_log_level,
            "root_dir": str(Path(self.root_dir).relative_to(self.bucket_name)),
            "gzip_content_types": self.gzip_content_types,
        }

    def get_fsspec_storage_options(self):
        s3_additional_kwargs = {}
        if self.default_acl:
            s3_additional_kwargs["ACL"] = self.default_acl
        if self.encryption:
            s3_additional_kwargs["ServerSideEncryption"] = "AES256"
        if self.reduced_redundancy:
            s3_additional_kwargs["StorageClass"] = "REDUCED_REDUNDANCY"

        return {
            "anon": not self.access_key and not self.security_token,
            "key": self.access_key,
            "secret": self.secret_key,
            "token": self.security_token,
            "use_ssl": self.use_ssl,
            "s3_additional_kwargs": s3_additional_kwargs,
            "client_kwargs": {
                "endpoint_url": self.endpoint_url,
            },
        }

    def _strip_signing_parameters(self, url):
        """Duplicated Unsiged URLs from Django-Stroage

        Method from: https://github.com/jschneier/django-storages/blob/master/storages/backends/s3boto3.py

        Boto3 does not currently support generating URLs that are unsigned. Instead we
        take the signed URLs and strip any querystring params related to signing and expiration.
        Note that this may end up with URLs that are still invalid, especially if params are
        passed in that only work with signed URLs, e.g. response header params.
        The code attempts to strip all query parameters that match names of known parameters
        from v2 and v4 signatures, regardless of the actual signature version used.
        """
        split_url = urlsplit(url)
        qs = parse_qsl(split_url.query, keep_blank_values=True)
        blacklist = {
            "x-amz-algorithm",
            "x-amz-credential",
            "x-amz-date",
            "x-amz-expires",
            "x-amz-signedheaders",
            "x-amz-signature",
            "x-amz-security-token",
            "awsaccesskeyid",
            "expires",
            "signature",
        }
        filtered_qs = ((key, val) for key, val in qs if key.lower() not in blacklist)
        # Note: Parameters that did not have a value in the original query string will have
        # an '=' sign appended to it, e.g ?foo&bar becomes ?foo=&bar=
        joined_qs = ("=".join(keyval) for keyval in filtered_qs)
        split_url = split_url._replace(query="&".join(joined_qs))
        return split_url.geturl()

    def url(self, object_name, parameters=None, expire=None):
        """Return Pre-signed URL

        Download URL to `object_name` in the connected bucket with a
        fixed expire time

        Documentation
        -------------
        https://boto3.amazonaws.com/v1/documentation/api/latest/guide/s3-presigned-urls.html


        Parameters
        ----------
        :param object_name: 'key' or name of object in bucket
        :type  object_name: str

        :param parameters: Dictionary of parameters to send to the method (BOTO3)
        :type  parameters: dict

        :param expire: Time in seconds for the presigned URL to remain valid
        :type  expire: int

        :return: Presigned URL as string. If error, returns None.
        :rtype str
        """
        params = parameters.copy() if parameters else {}
        params["Bucket"] = self.bucket_name
        params["Key"] = self.fs._join(object_name).split("/", 1)[
            -1
        ]  # strip the bucket name

        if expire is None:
            expire = self.querystring_expire

        url = sync(
            self.fs.fs.loop,
            self.fs.fs.s3.generate_presigned_url,
            "get_object",
            Params=params,
            ExpiresIn=expire,
        )

        if self.querystring_auth:
            return url
        else:
            return self._strip_signing_parameters(url)

    def get_storage_url(self, filename=None, suffix="tar.gz", encode_params=True):
        filename = (
            filename if filename is not None else self._get_unique_filename(suffix)
        )

        params = {}
        if encode_params:
            if self.default_acl:
                params["acl"] = self.default_acl

            if self.access_key:
                params["key"] = self.access_key

            if self.secret_key:
                params["secret"] = self.secret_key

            if self.security_token:
                params["token"] = self.security_token

            if self.endpoint_url:
                params["endpoint"] = self.endpoint_url

        return (
            filename,
            f"s3://{os.path.join(self.root_dir, filename)}{'?' if params else ''}{parse.urlencode(params) if params else ''}",
        )

    @contextlib.contextmanager
    def open(self, path, *args, **kwargs):
        if self.default_acl:
            kwargs.setdefault("acl", self.default_acl)

        with super().open(path, *args, **kwargs) as f:
            yield f



class OasisException(Exception):
    """
    Oasis base exception class

    Example
    -------
    In [call]: raise OasisException('Error Message 1', OSError('Root of error'))
    OasisException: Error Message 1, from OSError: Root of error
    """

    def __init__(self, msg, original_exception=None):
        self.original_exception = original_exception
        if original_exception:
            # This is wrapped execption
            super(OasisException, self).__init__(
                f"{msg}, {original_exception.__class__.__name__}: {original_exception}"
            )
        else:
            # Message only exception
            super(OasisException, self).__init__(msg)



import json
import os
from typing import Optional, Tuple, TypedDict, Union

from typing_extensions import NotRequired

from oasis_data_manager.config import ConfigError, load_class
from oasis_data_manager.filestore.backends.base import BaseStorage
from oasis_data_manager.filestore.backends.local import LocalStorage


class BaseStorageConfig(TypedDict):
    root_dir: str
    cache_dir: str


class LocalStorageConfig(BaseStorageConfig):
    pass


class S3StorageConfig(BaseStorageConfig):
    bucket_name: NotRequired[str]
    access_key: NotRequired[str]
    secret_key: NotRequired[str]
    endpoint_url: NotRequired[str]
    file_overwrite: NotRequired[bool]
    object_parameters: NotRequired[dict]
    auto_create_bucket: NotRequired[bool]
    default_acl: NotRequired[str]
    bucket_acl: NotRequired[str]
    querystring_auth: NotRequired[bool]
    querystring_expire: NotRequired[int]
    reduced_redundancy: NotRequired[bool]
    location: NotRequired[str]
    encryption: NotRequired[bool]
    security_token: NotRequired[str]
    secure_urls: NotRequired[bool]
    file_name_charset: NotRequired[str]
    gzip: NotRequired[bool]
    preload_metadata: NotRequired[bool]
    url_protocol: NotRequired[str]
    region_name: NotRequired[str]
    use_ssl: NotRequired[str]
    verify: NotRequired[bool]
    max_memory_size: NotRequired[int]
    shared_bucket: NotRequired[bool]
    aws_log_level: NotRequired[str]
    gzip_content_types: NotRequired[Tuple[str, ...]]


class AbfsStorageConfig(BaseStorageConfig):
    account_name: NotRequired[str]
    account_key: NotRequired[str]
    azure_container: NotRequired[str]
    location: NotRequired[str]
    connection_string: NotRequired[str]
    shared_container: NotRequired[bool]
    azure_ssl: NotRequired[bool]
    upload_max_conn: NotRequired[int]
    timeout: NotRequired[int]
    max_memory_size: NotRequired[int]
    expiration_secs: NotRequired[int]
    overwrite_files: NotRequired[bool]
    default_content_type: NotRequired[str]
    cache_control: NotRequired[str]
    sas_token: NotRequired[str]
    custom_domain: NotRequired[str]
    token_credential: NotRequired[str]
    azure_log_level: NotRequired[int]
    endpoint_url: NotRequired[str]


class StorageConfig(TypedDict):
    storage_class: str
    options: Union[
        LocalStorageConfig,
        S3StorageConfig,
        AbfsStorageConfig,
    ]


def get_storage_from_config(config: StorageConfig):
    cls = load_class(config["storage_class"], BaseStorage)
    return cls(**config["options"])


def get_storage_from_config_path(config_path: Optional[str], fallback_path: str):
    """
    Loads the config from the supplied path. If no config path is provided or the path
    doesn't exist a local file store object will be created with the root set to the
    fallback path.

    :param config_path: The path to the config file to load
    :param fallback_path: The path for the local file store should the config path not exist
    """
    if config_path and os.path.exists(config_path):
        with open(config_path) as f:
            config: StorageConfig = json.load(f)
            model_storage = get_storage_from_config(config)
    elif fallback_path:
        model_storage = LocalStorage(
            root_dir=fallback_path,
        )
    else:
        raise ConfigError(
            "The given config path does not exist and no fallback path was given to create the local storage from"
        )

    return model_storage



"""
    Readers to replace direct usage of pd.read_csv/read_parquet and allows for filters() & sql()
    to be provided.
"""

import io
import logging
import pathlib
from typing import Iterable

import dask
import dask_geopandas as dgpd
import geopandas as gpd
import pandas as pd
from dask import dataframe as dd
from dask_sql import Context
from dask_sql.utils import ParsingException
from distributed import Client

from ..filestore.backends.base import BaseStorage
from .exceptions import InvalidSQLException

dask.config.set(
    {"dataframe.convert-string": False}
)  # allows dask sql to support pyarrow
logger = logging.getLogger("oasis_data_manager.df_reader.reader")


class OasisReader:
    """
    Base reader.

    as_pandas(), sql() & filter() can all be chained with self.has_read controlling whether the base
    read (read_csv/read_parquet) needs to be triggered. This is because in the case of spark
    we need to read differently depending on if the intention is to do sql or filter.
    """

    def __init__(
        self,
        filename_or_buffer,
        storage: BaseStorage,
        *args,
        dataframe=None,
        has_read=False,
        **kwargs,
    ):
        self.filename_or_buffer = filename_or_buffer
        self.storage = storage
        self._df = dataframe
        self.has_read = has_read
        self.reader_args = args
        self.reader_kwargs = kwargs

        if not filename_or_buffer:
            if dataframe is None and not has_read:
                raise RuntimeError(
                    "Reader must be initialised with either a "
                    "filename_or_buffer or by passing a dataframe "
                    "and has_read=True"
                )
            else:
                self.read_from_dataframe()

        if (
            filename_or_buffer
            and isinstance(self.filename_or_buffer, str)
            and self.filename_or_buffer.lower().endswith(".zip")
        ):
            self.reader_kwargs["compression"] = "zip"

    @property
    def df(self):
        self._read()
        return self._df

    @df.setter
    def df(self, other):
        self._df = other

    def read_csv(self, *args, **kwargs):
        raise NotImplementedError()

    def read_parquet(self, *args, **kwargs):
        raise NotImplementedError()

    def _read(self):
        if not self.has_read:
            if hasattr(self.filename_or_buffer, "name"):
                extension = pathlib.Path(self.filename_or_buffer.name).suffix
            else:
                extension = pathlib.Path(self.filename_or_buffer).suffix

            if extension in [".parquet", ".pq"]:
                self.has_read = True
                self.read_parquet(*self.reader_args, **self.reader_kwargs)
            else:
                # assume the file is csv if not parquet
                self.has_read = True
                self.read_csv(*self.reader_args, **self.reader_kwargs)

        return self

    def copy_with_df(self, df):
        return type(self)(
            self.filename_or_buffer, self.storage, dataframe=df, has_read=self.has_read
        )

    def filter(self, filters):
        self._read()

        df = self.df
        for df_filter in filters if isinstance(filters, Iterable) else [filters]:
            df = df_filter(df)

        return self.copy_with_df(df)

    def sql(self, sql):
        if sql:
            self._read()
            return self.apply_sql(sql)
        return self

    def query(self, fn):
        return fn(self.df)

    def as_pandas(self):
        self._read()
        return self.df

    def read_from_dataframe(self):
        pass


class OasisPandasReader(OasisReader):
    def read_csv(self, *args, **kwargs):
        if isinstance(self.filename_or_buffer, str):
            if self.filename_or_buffer.startswith(
                "http://"
            ) or self.filename_or_buffer.startswith("https://"):
                self.df = pd.read_csv(self.filename_or_buffer, *args, **kwargs)
            else:
                _, uri = self.storage.get_storage_url(
                    self.filename_or_buffer, encode_params=False
                )
                self.df = pd.read_csv(
                    uri,
                    *args,
                    **kwargs,
                    storage_options=self.storage.get_fsspec_storage_options(),
                )
        else:
            self.df = pd.read_csv(self.filename_or_buffer, *args, **kwargs)

    def read_parquet(self, *args, **kwargs):
        if isinstance(self.filename_or_buffer, str):
            if self.filename_or_buffer.startswith(
                "http://"
            ) or self.filename_or_buffer.startswith("https://"):
                self.df = pd.read_parquet(self.filename_or_buffer, *args, **kwargs)
            else:
                _, uri = self.storage.get_storage_url(
                    self.filename_or_buffer, encode_params=False
                )
                self.df = pd.read_parquet(
                    uri,
                    *args,
                    **kwargs,
                    storage_options=self.storage.get_fsspec_storage_options(),
                )
        else:
            self.df = pd.read_parquet(self.filename_or_buffer, *args, **kwargs)

    def apply_geo(self, shape_filename_path, *args, drop_geo=True, **kwargs):
        """
        Read in a shape file and return the _read file with geo data joined.
        """
        # TODO: fix this so that it can work with non local files
        # with self.storage.open(self.shape_filename_path) as f:
        #     shape_df = gpd.read_file(f)

        shape_df = gpd.read_file(shape_filename_path)

        # for situations where the columns in the source data are different.
        lon_col = kwargs.get("geo_lon_col", "longitude")
        lat_col = kwargs.get("geo_lat_col", "latitude")

        df_columns = self.df.columns.tolist()
        if lat_col not in df_columns or lon_col not in df_columns:
            logger.warning("Invalid shape file provided")
            # temp until we decide on handling, i.e don't return full data if it fails.
            return self.copy_with_df(pd.DataFrame.from_dict({}))

        # convert read df to geo
        df = gpd.GeoDataFrame(
            self.df, geometry=gpd.points_from_xy(self.df[lon_col], self.df[lat_col])
        )

        # Make sure they're using the same projection reference
        df.crs = shape_df.crs

        # join the datasets, matching `geometry` to points within the shape df
        df = df.sjoin(shape_df, how="inner")

        if drop_geo:
            df = df.drop(shape_df.columns.tolist() + ["index_right"], axis=1)

        return self.copy_with_df(df)


class OasisPandasReaderCSV(OasisPandasReader):
    pass


class OasisPandasReaderParquet(OasisPandasReader):
    pass


class OasisDaskReader(OasisReader):
    sql_table_name = "table"

    def __init__(self, *args, client_address=None, **kwargs):
        if client_address:
            self.client = Client(client_address, set_as_default=False)
        else:
            self.client = None

        self.sql_context = Context()
        self.table_names = [self.sql_table_name]
        self.pre_sql_columns = []

        super().__init__(*args, **kwargs)

    def copy_with_df(self, df):
        res = super().copy_with_df(df)
        res.client = self.client
        return res

    def apply_geo(self, shape_filename_path, *args, drop_geo=True, **kwargs):
        """
        Read in a shape file and return the _read file with geo data joined.
        """
        # TODO: fix this so that it can work with non local files
        # with self.storage.open(self.shape_filename_path) as f:
        #     shape_df = dgpd.read_file(f, npartitions=1)

        shape_df = dgpd.read_file(shape_filename_path, npartitions=1)

        # for situations where the columns in the source data are different.
        lon_col = kwargs.get("geo_lon_col", "longitude")
        lat_col = kwargs.get("geo_lat_col", "latitude")

        df_columns = self.df.columns.tolist()
        if lat_col not in df_columns or lon_col not in df_columns:
            logger.warning("Invalid shape file provided")
            # temp until we decide on handling, i.e don't return full data if it fails.
            return self.copy_with_df(dd.DataFrame.from_dict({}, npartitions=1))

        df = self.df.copy()

        # convert read df to geo
        df["geometry"] = dgpd.points_from_xy(df, lon_col, lat_col)
        df = dgpd.from_dask_dataframe(df)

        # Make sure they're using the same projection reference
        df.crs = shape_df.crs

        # join the datasets, matching `geometry` to points within the shape df
        df = df.sjoin(shape_df, how="inner")

        if drop_geo:
            df = df.drop(shape_df.columns.tolist() + ["index_right"], axis=1)

        return self.copy_with_df(df)

    def apply_sql(self, sql):
        df = self.df.copy()
        try:
            # Initially this was the filename, but some filenames are invalid for the table,
            # is it ok to call it the same name all the time? Mapped to DaskDataTable in case
            # we need to change this.
            self.sql_context.create_table("DaskDataTable", self.df)
            formatted_sql = sql.replace(self.sql_table_name, "DaskDataTable")

            self.pre_sql_columns.extend(df.columns)

            # dask expects the columns to be lower case, which won't match some data
            df = self.sql_context.sql(
                formatted_sql,
                config_options={"sql.identifier.case_sensitive": False},
            )
            # which means we then need to map the columns back to the original
            # and allow for any aggregations to be retained
            validated_columns = []
            for v in df.columns:
                pre = False
                for x in self.pre_sql_columns:
                    if v.lower() == x.lower():
                        validated_columns.append(x)
                        pre = True

                if not pre:
                    validated_columns.append(v)
            df.columns = validated_columns

            return self.copy_with_df(df)
        except ParsingException:
            raise InvalidSQLException

    def join(self, df, table_name):
        """
        Creates a secondary table as a sql table in order to allow joins when apply_sql is called.
        """
        if table_name in self.table_names:
            raise RuntimeError(
                f"Table name already in use: [{','.join(self.table_names)}]"
            )
        self.pre_sql_columns.extend(df.columns)
        self.sql_context.create_table(table_name, df)
        self.table_names.append(table_name)
        return self

    def read_from_dataframe(self):
        if not isinstance(self.df, dd.DataFrame):
            self.df = dd.from_pandas(self.df, npartitions=1)

    def as_pandas(self):
        super().as_pandas()
        if self.client:
            return self.client.compute(self.df).result()
        else:
            return self.df.compute()

    def read_dict(self, data):
        self.df = dd.DataFrame.from_dict(data)

    def read_csv(self, *args, **kwargs):
        # remove standard pandas kwargs which will case an issue in dask.
        dask_safe_kwargs = kwargs.copy()
        dask_safe_kwargs.pop("memory_map", None)
        dask_safe_kwargs.pop("low_memory", None)

        filename_or_buffer = self.filename_or_buffer
        if isinstance(filename_or_buffer, pathlib.PosixPath):
            filename_or_buffer = str(self.filename_or_buffer)

        if isinstance(filename_or_buffer, io.TextIOWrapper) or isinstance(
            filename_or_buffer, io.BufferedReader
        ):
            filename_or_buffer = filename_or_buffer.name

        # django files
        if hasattr(filename_or_buffer, "path"):
            filename_or_buffer = filename_or_buffer.path

        _, uri = self.storage.get_storage_url(filename_or_buffer, encode_params=False)
        self.df = dd.read_csv(
            uri,
            *args,
            **dask_safe_kwargs,
            storage_options=self.storage.get_fsspec_storage_options(),
        )

    def read_parquet(self, *args, **kwargs):
        if isinstance(self.filename_or_buffer, str):
            _, uri = self.storage.get_storage_url(
                self.filename_or_buffer, encode_params=False
            )
            filename = uri
            kwargs["storage_options"] = self.storage.get_fsspec_storage_options()
        else:
            filename = self.filename_or_buffer

        self.df = dd.read_parquet(
            filename,
            *args,
            **kwargs,
        )

        # dask-sql doesn't handle categorical columns, but we need to be careful
        # how we convert them, if an assign is used we will end up stopping
        # the `Predicate pushdown optimization` within dask-sql from applying the
        # sql to the read_parquet filters.
        categories_to_convert = {}
        for col in self.df.select_dtypes(include="category").columns:
            categories_to_convert[col] = self.df[col].dtype.categories.dtype
        self.df = self.df.astype(categories_to_convert)


class OasisDaskReaderCSV(OasisDaskReader):
    pass


class OasisDaskReaderParquet(OasisDaskReader):
    pass



import contextlib
import urllib.parse
from urllib import parse

import fsspec


def split_s3_url(parts):
    query = parse.parse_qs(parts.query)
    params = {
        "acl": query.get("acl", [None])[0],
        "anon": "key" not in query and "token" not in query,
        "key": query.get("key", [None])[0],
        "secret": query.get("secret", [None])[0],
        "token": query.get("token", [None])[0],
        "use_ssl": query.get("ssl", ["True"])[0] == "True",
        "client_kwargs": {
            "endpoint_url": query.get("endpoint", [None])[0],
        },
    }

    return (
        urllib.parse.urlunparse(
            (
                parts[0],
                parts[1],
                parts[2],
                parts[3],
                "",
                parts[5],
            )
        ),
        params,
    )


def split_azure_url(parts):
    connection_string = ""

    query = parse.parse_qs(parts.query)
    if "connection_string" in query:
        connection_string = parts.get("connection_string")[0]
    else:
        if "endpoint" in query:
            connection_string += f"BlobEndpoint={ query.get('endpoint', [None])[0]};"
        if "account" in query:
            connection_string += f"AccountName={ query.get('account', [None])[0] };"
        if "key" in query:
            connection_string += f"AccountKey={ query.get('key', [None])[0] };"

    params = {
        "anon": "key" not in query and "secret" not in query,
        "connection_string": connection_string,
        "client_id": query.get("client", [None])[0],
        "client_secret": query.get("secret", [None])[0],
        "tenant_id": query.get("tenant", [None])[0],
        "account_name": query.get("account", [None])[0],
        "account_key": query.get("key", [None])[0],
        "use_ssl": query.get("ssl", ["True"])[0] == "True",
    }

    return (
        urllib.parse.urlunparse(
            (
                parts[0],
                parts[1],
                parts[2],
                parts[3],
                "",
                parts[5],
            )
        ),
        params,
    )


def parse_url_options(path):
    opener = fsspec.open

    params = {}

    url_parts = parse.urlparse(path)
    if url_parts.scheme == "s3":
        path, params = split_s3_url(url_parts)
    elif url_parts.scheme in ["az", "abfs", "adl"]:
        path, params = split_azure_url(url_parts)

    return opener, path, params


@contextlib.contextmanager
def open_file(path: str, mode, **kwargs):
    opener, path, opts = parse_url_options(path)
    with opener(path, mode, **kwargs, **opts) as f:
        yield f



import contextlib
from pathlib import Path

from oasis_data_manager.filestore.backends.base import BaseStorage, MissingInputsException


class LocalStorage(BaseStorage):
    """
    Implements storage for a local filesystem. All paths passed to the
    storage should be relative to the media root.
    """

    storage_connector = "FS-SHARE"
    fsspec_filesystem_class = None

    @property
    def config_options(self):
        return {
            "root_dir": self.root_dir,
        }

    def get_storage_url(self, filename=None, suffix="tar.gz", **kwargs):
        filename = (
            filename if filename is not None else self._get_unique_filename(suffix)
        )
        return filename, f"file://{Path(self.root_dir, filename)}"

    def get_fsspec_storage_options(self):
        return {}

    @contextlib.contextmanager
    def with_fileno(self, path, mode="rb"):
        with self.open(path, mode) as f:
            yield f

    def get_from_cache(self, reference, required=False, no_cache_target=None):
        # null ref given
        if not reference:
            if required:
                raise MissingInputsException(reference)
            else:
                return None

        return self.fs._join(reference)



import os
import re

from setuptools import find_packages, setup

SCRIPT_DIR = os.path.abspath(os.path.dirname(__file__))


def get_version():
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    with open(
        os.path.join(SCRIPT_DIR, "oasis_data_manager", "__init__.py"), encoding="utf-8"
    ) as init_py:
        return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py.read()).group(1)


def get_install_requirements():
    with open(
        os.path.join(SCRIPT_DIR, "requirements-package.in"), encoding="utf-8"
    ) as reqs:
        return reqs.readlines()


version = get_version()


setup(
    name="oasis-data-manager",
    version=version,
    packages=find_packages(exclude=("tests", "tests.*", "tests.*.*")),
    include_package_data=True,
    package_data={
        "": [
            "LICENSE",
        ],
    },
    exclude_package_data={
        "": ["__pycache__", "*.py[co]"],
    },
    license="BSD 3-Clause",
    description="",
    long_description="",
    long_description_content_type="text/markdown",
    url="https://github.com/OasisLMF/OasisDataManager",
    author="Oasis LMF",
    author_email="support@oasislmf.org",
    keywords="",
    python_requires=">=3.6",
    install_requires=get_install_requirements(),
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
    ],
)



import logging
import os
from pathlib import Path
from typing import Optional
from urllib import parse

import fsspec

from ..log import set_azure_log_level
from .base import BaseStorage


class AzureABFSStorage(BaseStorage):
    # https://docs.microsoft.com/en-us/azure/storage/blobs/storage-quickstart-blobs-python
    fsspec_filesystem_class = fsspec.get_filesystem_class("abfs")

    def __init__(
        self,
        account_name: Optional[str] = None,
        account_key: Optional[str] = None,
        azure_container: Optional[str] = None,
        location="",
        connection_string: Optional[str] = None,
        shared_container=True,
        azure_ssl=True,
        upload_max_conn=2,
        timeout=20,
        max_memory_size=2 * 1024 * 1024,
        expiration_secs: Optional[int] = None,
        overwrite_files=True,
        default_content_type="application/octet-stream",
        cache_control: Optional[str] = None,
        sas_token: Optional[str] = None,
        custom_domain: Optional[str] = None,
        token_credential: Optional[str] = None,
        azure_log_level=logging.ERROR,
        root_dir="",
        endpoint_url: Optional[str] = None,
        **kwargs,
    ):
        self._service_client = None
        self._client = None

        # Required
        self.account_name = account_name
        self.account_key = account_key
        self.azure_container = azure_container

        # Optional
        self._connection_string = connection_string
        self.shared_container = shared_container
        self.azure_ssl = azure_ssl
        self.upload_max_conn = upload_max_conn
        self.timeout = timeout
        self.max_memory_size = max_memory_size
        self.expiration_secs = expiration_secs
        self.overwrite_files = overwrite_files
        self.default_content_type = default_content_type
        self.cache_control = cache_control
        self.sas_token = sas_token
        self.custom_domain = custom_domain
        self.token_credential = token_credential
        self.azure_log_level = azure_log_level
        self.azure_protocol = "https" if self.azure_ssl else "http"
        self.endpoint_url = endpoint_url
        set_azure_log_level(self.azure_log_level)

        root_dir = os.path.join(self.azure_container or "", root_dir or location or "")
        if root_dir.startswith(os.path.sep):
            root_dir = root_dir[1:]
        if root_dir.endswith(os.path.sep):
            root_dir = root_dir[:-1]

        super(AzureABFSStorage, self).__init__(root_dir=root_dir, **kwargs)

    @property
    def config_options(self):
        return {
            "account_name": self.account_name,
            "account_key": self.account_key,
            "azure_container": self.azure_container,
            "connection_string": self.connection_string,
            "shared_container": self.shared_container,
            "azure_ssl": self.azure_ssl,
            "upload_max_conn": self.upload_max_conn,
            "timeout": self.timeout,
            "max_memory_size": self.max_memory_size,
            "expiration_secs": self.expiration_secs,
            "overwrite_files": self.overwrite_files,
            "default_content_type": self.default_content_type,
            "cache_control": self.cache_control,
            "sas_token": self.sas_token,
            "custom_domain": self.custom_domain,
            "token_credential": self.token_credential,
            "azure_log_level": self.azure_log_level,
            "root_dir": str(Path(self.root_dir).relative_to(self.azure_container)),
            "endpoint_url": self.endpoint_url,
        }

    def url(self, object_name, parameters=None, expire=None):
        blob_key = self.fs._join(object_name)
        return self.fs.fs.url(blob_key)

    def get_fsspec_storage_options(self):
        return {
            "anon": not self.account_key,
            "connection_string": self.connection_string,
            "account_name": self.account_name,
            "account_key": self.account_key,
            "use_ssl": self.azure_ssl,
        }

    @property
    def connection_string(self):
        if self._connection_string:
            return self._connection_string
        else:
            cs = ""
            if self.endpoint_url:
                cs += f"BlobEndpoint={self.endpoint_url};"
            if self.account_name:
                cs += f"AccountName={self.account_name};"
            if self.account_key:
                cs += f"AccountKey={self.account_key};"

            return cs

    def get_storage_url(self, filename=None, suffix="tar.gz", encode_params=True):
        filename = (
            filename if filename is not None else self._get_unique_filename(suffix)
        )

        params = {}
        if encode_params:
            if self.connection_string:
                params["connection_string"] = self.connection_string
            else:
                if self.account_name:
                    params["account"] = self.account_name

                if self.account_key:
                    params["key"] = self.account_key

                if self.endpoint_url:
                    params["endpoint"] = self.endpoint_url

        return (
            filename,
            f"abfs://{os.path.join(self.root_dir, filename)}{'?' if params else ''}{parse.urlencode(params) if params else ''}",
        )



#!/usr/bin/env python

import h5py
import pandas as pd

from oasis_data_manager.complex.complex import Adjustment, FileStoreComplexData
from oasis_data_manager.df_reader.reader import OasisDaskReader
from oasis_data_manager.filestore.backends.local import LocalStorage


class AddColAdjustment(Adjustment):
    @classmethod
    def apply(cls, df):
        df["else"] = "test"
        return df


class FloodDataExample(FileStoreComplexData):
    # Note this file needs to be sourced. TODO - is this public? If not is there a public/file we can include?
    filename = "tropical_cyclone_10synth_tracks_150arcsec_rcp26_KNA_2080.hdf5"
    adjustments = [AddColAdjustment]
    sql = "SELECT * FROM table WHERE event_id > 3000"

    def to_dataframe(self, result) -> pd.DataFrame:
        result = h5py.File(result)

        df = pd.DataFrame(list(result["event_id"]), columns=["event_id"])
        for m in ["event_name", "date", "frequency", "orig"]:
            df[m] = list(result[m])
        df = df.reset_index()

        return df


if __name__ == "__main__":
    reader = OasisDaskReader
    test_storage = LocalStorage("/tmp")

    result = FloodDataExample(storage=test_storage).run()
    print(result.as_pandas())



import base64
import contextlib
import io
import logging
import os
import shutil
import tarfile
import tempfile
import uuid
from pathlib import Path
from typing import Optional, Tuple, Type, Union
from urllib.parse import urlparse
from urllib.request import urlopen

import fsspec
from fsspec.implementations.dirfs import DirFileSystem

from oasis_data_manager.errors import OasisException

LOG_FILE_SUFFIX = "txt"
ARCHIVE_FILE_SUFFIX = "tar.gz"


class MissingInputsException(OasisException):
    def __init__(self, input_filepath):
        super(MissingInputsException, self).__init__(
            "Input file not found: {}".format(input_filepath)
        )


class StrictRootDirFs(DirFileSystem):
    def _path_is_in_root(self, path):
        return os.path.abspath(path).startswith(os.path.abspath(self.path))

    def _join(self, path):
        res = super()._join(path)

        if isinstance(res, str):
            if not self._path_is_in_root(res):
                raise FileNotFoundError(path)
        else:
            for p in res:
                if not self._path_is_in_root(p):
                    raise FileNotFoundError(p)

        return res

    def exists(self, path):
        try:
            return super().exists(path)
        except FileNotFoundError:
            return False

    def isfile(self, path):
        try:
            return super().isfile(path)
        except FileNotFoundError:
            return False

    def isdir(self, path):
        try:
            return super().isdir(path)
        except FileNotFoundError:
            return False


class BaseStorage(object):
    """Base storage class

    Implements storage for a local fileshare between
    `server` and `worker` containers
    """

    storage_connector: str
    fsspec_filesystem_class: Optional[Type[fsspec.AbstractFileSystem]]

    def __init__(
        self, root_dir="", cache_dir: Union[str, None] = "/tmp/data-cache", logger=None
    ):
        # Use for caching files across multiple runs, set value 'None' or 'False' to disable
        self.cache_root = cache_dir
        self.root_dir = root_dir

        self.logger = logger or logging.getLogger()
        self._fs: Optional[StrictRootDirFs] = None

    def to_config(self) -> dict:
        return {
            "storage_class": f"{self.__module__}.{type(self).__name__}",
            "options": self.config_options,
        }

    @property
    def config_options(self):
        raise NotImplementedError()

    def _get_unique_filename(self, suffix=""):
        """Returns a unique name

        Parameters
        ----------
        :param suffix: Set the filename extension
        :type suffix: str

        :return: filename string
        :rtype str
        """
        return "{}.{}".format(uuid.uuid4().hex, suffix)

    def _is_valid_url(self, url):
        """Check if a String is a valid url

        Parameters
        ----------
        :param url: String to check
        :type  url: str

        :return: `True` if URL otherwise `False`
        :rtype boolean
        """
        if url:
            result = urlparse(url)
            return all([result.scheme, result.netloc]) and result.scheme in [
                "http",
                "https",
            ]
        else:
            return False

    def extract(self, archive_fp, directory, storage_subdir=""):
        """Extract tar file

        Parameters
        ----------
        :param archive_fp: Path to archive file
        :type  archive_fp: str

        :param directory: Path to extract contents to.
        :type  directory: str

        :param storage_subdir: Store object in given sub directory
        :type  storage_subdir: str
        """
        with tempfile.TemporaryDirectory() as temp_dir_path:
            local_archive_path = self.get(
                archive_fp,
                os.path.join(temp_dir_path, os.path.basename(archive_fp)),
                subdir=storage_subdir,
            )
            with tarfile.open(local_archive_path) as f:
                os.makedirs(directory, exist_ok=True)
                f.extractall(directory)

    def compress(self, archive_fp, directory, arcname=None):
        """Compress a directory

        Parameters
        ----------
        :param archive_fp: Path to archive file
        :type  archive_fp: str

        :param directory: Path to archive.
        :type  directory: str

        :param arcname: If given, `arcname' set an alternative
                        name for the file in the archive.
        :type arcname: str
        """
        arcname = arcname if arcname else "/"
        with tarfile.open(archive_fp, "w:gz") as tar:
            tar.add(directory, arcname=arcname)

    def get_from_cache(self, reference, required=False, no_cache_target=None):
        """
        Retrieves a file from the storage and stores it in the cache.
        If it already exists in te cache the fetching step will be skipped

        If URL: download the object and place in `output_dir`
        If Filename: return stored file path of the shared object

        Parameters
        ----------
        :param reference: Filename or download URL
        :type  reference: str

        :param no_cache_target: A path to store the file at if no cache root is set
        :type  no_cache_target: str

        :return: Absolute filepath to stored Object
        :rtype str
        """
        # null ref given
        if not reference:
            if required:
                raise MissingInputsException(reference)
            else:
                return None

        # check if the file is in the cache, if so return that path
        cache_filename = base64.b64encode(reference.encode()).decode()
        if self.cache_root:
            cached_file = os.path.join(self.cache_root, cache_filename)
        else:
            os.makedirs(os.path.dirname(no_cache_target), exist_ok=True)
            cached_file = no_cache_target

        if self.cache_root:
            if os.path.exists(cached_file):
                logging.info("Get from Cache: {}".format(reference))
                return cached_file

        if self._is_valid_url(reference):
            # if the file is not in the path, and is a url download it to the cache
            response = urlopen(reference)
            fdata = response.read()

            with io.open(cached_file, "w+b") as f:
                f.write(fdata)
                logging.info("Get from URL: {}".format(reference))
        else:
            # otherwise get it from the storage and add it to the cache
            self.fs.get(reference, cached_file, recursive=True)

        return cached_file

    def get(self, reference, output_path="", subdir="", required=False):
        """Retrieve stored object and stores it in the output path

        Top level 'get from storage' function
        Check if `reference` is either download `URL` or filename

        If URL: download the object and place in `output_dir`
        If Filename: return stored file path of the shared object

        Parameters
        ----------
        :param reference: Filename or download URL
        :type  reference: str

        :param output_path: If given, download to that directory.
        :type  output_path: str

        :param subdir: Store a file under this sub directory path
        :type  subdir: str

        :return: Absolute filepath to stored Object
        :rtype str
        """
        target = os.path.abspath(
            os.path.join(output_path, subdir) if subdir else output_path
        )

        if os.path.isdir(target):
            fname = reference
            if self._is_valid_url(reference):
                fname = os.path.basename(urlparse(reference).path)

            target = os.path.join(output_path, fname)

        res = self.get_from_cache(reference, required=required, no_cache_target=target)
        if res:
            if res != target:
                os.makedirs(os.path.dirname(target), exist_ok=True)
                shutil.copy(res, target)
            return target

        return None

    def put(self, reference, filename=None, subdir="", suffix=None, arcname=None):
        """Place object in storage

        Top level send to storage function,
        Create new connector classes by Overriding
        `self._store_file( .. )` and `self._store_dir( .. )`

        Parameters
        ----------
        :param reference: Path to either a `File` or `Directory`
        :type  reference: str

        :param filename: Set the name of stored file, instead of uuid
        :type  filename: str

        :param subdir: Store a file under this sub directory path
        :type  subdir: str

        :param arcname: If given, `arcname' set an alternative
                        name for the file in the archive.
        :type arcname: str

        :param suffix: Set the filename extension defaults to `tar.gz`
        :type suffix: str

        :return: access storage reference returned from self._store_file, self._store_dir
                 This will either be a pre-signed URL or absolute filepath
        :rtype str
        """
        if not reference:
            return None

        ext = "tar.gz" if not suffix else suffix
        filename = filename if filename else self._get_unique_filename(ext)
        storage_path = os.path.join(subdir, filename) if subdir else filename

        self.fs.mkdirs(os.path.dirname(storage_path), exist_ok=True)

        if os.path.isfile(reference):
            self.logger.info("Store file: {} -> {}".format(reference, storage_path))
            self.fs.put(reference, storage_path)
            return storage_path
        elif os.path.isdir(reference):
            self.logger.info("Store dir: {} -> {}".format(reference, storage_path))
            with tempfile.NamedTemporaryFile() as f:
                self.compress(f.name, reference, arcname)
                self.fs.put(f.name, storage_path)
            return storage_path
        else:
            return None

    def delete_file(self, reference):
        """
        Delete single file from shared storage

        :param reference: Path to `File`
        :type  reference: str
        """
        if self.fs.isfile(reference):
            self.fs.delete(reference)
            logging.info("Deleted Shared file: {}".format(reference))
        else:
            logging.info("Delete Error - Unknwon reference {}".format(reference))

    def delete_dir(self, reference):
        """
        Delete subdirectory from shared storage

        :param reference: Path to `Directory`
        :type  reference: str
        """
        if self.fs.isdir(reference):
            if Path("/") == Path(reference).resolve():
                logging.info("Delete Error - prevented media root deletion")
            else:
                self.fs.delete(reference, recursive=True)
                logging.info("Deleted shared dir: {}".format(reference))
        else:
            logging.info("Delete Error - Unknwon reference {}".format(reference))

    def create_traceback(self, stdout, stderr, output_dir=""):
        traceback_file = self._get_unique_filename(LOG_FILE_SUFFIX)
        with tempfile.NamedTemporaryFile("w") as f:
            if stdout:
                f.write(stdout)
            if stderr:
                f.write(stderr)

            self.put(f.name, filename=traceback_file)
        return traceback_file

    def get_storage_url(
        self, filename=None, suffix="tar.gz", encode_params=True
    ) -> Tuple[str, str]:
        raise NotImplementedError

    def get_fsspec_storage_options(self):
        return {}

    @property
    def fs(self) -> fsspec.AbstractFileSystem:
        if not self._fs:
            self._fs = StrictRootDirFs(
                path=self.root_dir,
                fs=(
                    self.fsspec_filesystem_class(**self.get_fsspec_storage_options())
                    if self.fsspec_filesystem_class
                    else None
                ),
            )
        return self._fs

    def exists(self, path):
        return self.fs.exists(path)

    def isfile(self, path):
        return self.fs.isfile(path)

    def listdir(self, path=""):
        return self.fs.listdir(path, detail=False)

    @contextlib.contextmanager
    def open(self, path, *args, **kwargs):
        if self._is_valid_url(path):
            with tempfile.TemporaryDirectory() as d:
                with open(
                    self.get_from_cache(path, no_cache_target=os.path.join(d, "f"))
                ) as f:
                    yield f
        else:
            with self.fs.open(path, *args, **kwargs) as f:
                yield f

    @contextlib.contextmanager
    def with_fileno(self, path, mode="rb"):
        with tempfile.TemporaryDirectory() as d:
            target = os.path.join(d, "fileno")
            path = self.get_from_cache(path, no_cache_target=target)

            with open(path, mode) as f:
                yield f
